#!/usr/bin/env python

import os, os.path, shutil, textwrap, sys, locale
import gtk, pango
import bzrlib, bzrlib.plugin, bzrlib.workingtree
import bzrlib.plugins.gtk.commit as gtk_commit
import bzrlib.plugins.gtk.diff as gtk_diff
import pygg

bzrlib.plugin.load_plugins()

class CategorySet (object) :
    def __init__ (self, *cats) :
        self._catname = {}
        self._catitem = {}
        self._cat = []
        for item, name in cats :
            self.add(item, name)
    def default (self) :
        return self._cat[0][0]
    def add (self, item, name) :
        if item in self._catname :
            raise ValueError, "duplicate category item %r" % item
        elif name in self._catitem :
            raise ValueError, "duplicate category name %r" % name
        if len(item) != 1 :
            raise ValueError, "expected a single character, got %r" % item
        self._catname[item] = name
        self._catitem[name] = item
        self._cat.append((item, name))
    def remove (self, key) :
        item, name = self[key]
        del self._catitem[name]
        del self._catname[item]
        self._cat.remove((item, name))
    def __iter__ (self) :
        return iter(self._cat)
    def __contains__ (self, key) :
        return (key in self._catname) or (key in self._catitem)
    def __getitem__ (self, key) :
        if key in self._catname :
            return key, self._catname[key]
        elif key in self._catitem :
            return self._catitem[key], key
        else :
            raise KeyError, "unknown category %r" % key

class ItemList (object) :
    def __init__ (self, categories, rev=False) :
        self._cat = categories
        self._entries = []
        self._rev = rev
    def add_item (self, cat, text) :
        cat, _ = self._cat[cat]
        self._entries.append((cat, text))
    def add_text (self, text) :
        self._entries.append(text)
    def __iter__ (self) :
        for e in self._entries :
            if isinstance(e, tuple) :
                yield e
    def _format_entry (self, item, text) :
        return "\n".join(textwrap.wrap("%s %s" % (item, text),
                                       initial_indent=" ",
                                       subsequent_indent="   "))
    def format (self) :
        parts = []
        for e in self._entries :
            try :
                i, t = e
                parts.append(self._format_entry(i, t))
            except ValueError :
                parts.append(e)
        if self._rev :
            parts.reverse()
        return "\n".join(parts)
    def _flush (self, parts) :
        if len(parts) > 0 :
            i, t = " ".join(parts).lstrip().split(" ", 1)
            self.add_item(i.strip(), t.strip())
    def load (self, path) :
        infile = open(path)
        lines = infile.readlines()
        infile.close()
        if self._rev :
            lines.reverse()
        items = set(" %s " % i for i, n in self._cat)
        parts = []
        for line in lines :
            if line[:3] in items :
                self._flush(parts)
                parts = [line]
            elif line[:3] == "   " :
                parts.append(line)
            else :
                self._flush(parts)
                parts = []
                self.add_text(line.strip())
        self._flush(parts)
    def save (self, path) :
        outfile = open(path, "w")
        print >>outfile, self.format()
        outfile.close()

encoding = locale.getpreferredencoding()

def _ (text) :
    global encoding
    try :
        return unicode(text, encoding).encode("utf8", "replace")
    except :
        return text

class FileManager (gtk.TreeView) :
    def __init__ (self) :
        self._files_map = ["name", "directory", "status", "path", "id"]
        self._files = gtk.TreeStore(str, bool, str, str, str)
        gtk.TreeView.__init__(self, self._files)
        self._celltxt = gtk.CellRendererText()
        self._celltxt.set_property("xalign", 0)
        self._cellpix = gtk.CellRendererPixbuf()
        self._col = gtk.TreeViewColumn(None)
        self._col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        self._col.pack_end(self._celltxt, True)
        self._col.add_attribute(self._celltxt, 'text', 0)
        self._col.pack_start(self._cellpix, False)
        self._col.set_cell_data_func(self._cellpix, self._file2pix)
        self.append_column(self._col)
        self.set_fixed_height_mode(True)
        self.set_headers_visible(False)
    _status2pix = {"removed" : gtk.STOCK_DELETE,
                   "added" : gtk.STOCK_NEW,
                   "renamed" : gtk.STOCK_COPY,
                   "renamed and modified" : gtk.STOCK_EDIT,
                   "modified" : gtk.STOCK_EDIT,
                   "unchanged" : gtk.STOCK_SAVE,
                   "?" : gtk.STOCK_DIALOG_QUESTION}
    def _file2pix (self, tree, cell, model, iter) :
        status = model.get(iter, 2)[0]
        cell.set_property("stock-id", self._status2pix[status])
        cell.set_property("stock-size", gtk.ICON_SIZE_MENU)
    def _load_dir (self, root) :
        for base, dirs, files in os.walk(root) :
            for name in sorted(dirs) :
                path = os.path.join(base, name)
                if name.startswith(".") :
                    continue
                new = self._files.append(self._dirs[base],
                                         (name, True, "?", path, None))
                self._dirs[path] = new
            for name in sorted(files) :
                path = os.path.join(base, name)
                if name.startswith(".") :
                    continue
                self._files.append(self._dirs[base],
                                   (name, False, "?", path, None))
    def load (self, wt, all) :
        try :
            wt.lock_read()
            bt = wt.basis_tree()
            bt.lock_read()
            content = tuple(wt.list_files(True))
            status = dict((ident, change)
                          for ident, _, change, _ in
                          gtk_diff._iter_changes_to_status(bt, wt))
        finally :
            bt.unlock()
            wt.unlock()
        self._files.clear()
        self._dirs = {}
        for path, cls, kind, ident, entry in content :
            if cls == "I" :
                continue
            elif cls == "?" and not all :
                continue
            path = wt.abspath(path).rstrip(os.sep)
            name = os.path.basename(path)
            if name.startswith(".") :
                continue
            parent = self._dirs.get(os.path.dirname(path),
                                    self._files.get_iter_root())
            if cls == "?" :
                _status = "?"
            else :
                _status = status.get(ident, "unchanged")
            new = self._files.append(parent,
                                     (name, kind == "directory",
                                      _status, path, ident))
            if kind == "directory" :
                self._dirs[path] = new
                if cls == "?" :
                    self._load_dir(path)
        self.expand_all()
    def get_file (self) :
        sel = self.get_selection()
        if sel.count_selected_rows() == 1 :
            _, paths = sel.get_selected_rows()
            return self._files.get(self._files.get_iter(paths[0]), 3, 1, 2, 4)

class TodoList (gtk.TreeView) :
    def __init__ (self) :
        self._todo_map = ["item", "text"]
        self._todo = gtk.ListStore(str, str)
        gtk.TreeView.__init__(self, self._todo)
        self.set_headers_visible(True)
        # category column
        self._cats = gtk.ListStore(str)
        self._cellcat = gtk.CellRendererCombo()
        self._cellcat.set_property("xalign", 0)
        self._cellcat.set_property("model", self._cats)
        self._cellcat.set_property("text-column", 0)
        self._cellcat.set_property("editable", True)
        self._cellcat.set_property("font", "monospace")
        self._cellcat.connect("edited", self._update_cat)
        col = gtk.TreeViewColumn("Categ.", self._cellcat)
        col.add_attribute(self._cellcat, "text", 0)
        self.append_column(col)
        # tasks column
        self._celltxt = gtk.CellRendererText()
        self._celltxt.set_property("xalign", 0)
        self._celltxt.set_property("editable", True)
        self._celltxt.set_property("font", "monospace")
        self._celltxt.connect("edited", self._update_task)
        col = gtk.TreeViewColumn("Task", self._celltxt)
        col.add_attribute(self._celltxt, "text", 1)
        self.append_column(col)
    def _update_cat (self, cell, path, text) :
        text = text.strip()
        if text in self._categories :
            self._todo.set(self._todo.get_iter(path), 0, text[0])
    def _update_task (self, cell, path, text) :
        self._todo.set(self._todo.get_iter(path), 1, text.strip())
    def add_item (self, category, text) :
        item, _ = self._categories[category]
        self._todo.append((item, text))
    def get_item (self) :
        sel = self.get_selection()
        if sel.count_selected_rows() == 1 :
            _, paths = sel.get_selected_rows()
            return self._todo.get(self._todo.get_iter(paths[0]), 0, 1)
    def del_item (self) :
        sel = self.get_selection()
        if sel.count_selected_rows() == 1 :
            _, paths = sel.get_selected_rows()
            return self._todo.remove(self._todo.get_iter(paths[0]))
    def get_all (self) :
        it = self._todo.get_iter_first()
        while it is not None :
            yield self._todo.get(it, 0, 1)
            it = self._todo.iter_next(it)
    def load (self, categories) :
        self._categories = categories
        self._cats.clear()
        for cat in categories :
            self._cats.append(("\t".join(cat),))
        todo = ItemList(categories)
        try :
            todo.load("TODO")
        except :
            return
        for item, text in todo :
            self.add_item(item, text)
    def save (self) :
        todo = ItemList(self._categories)
        todo.add_text("To do:")
        for item, text in self.get_all() :
            todo.add_item(item, text)
        todo.save("TODO")

class CategoryTextView (gtk.TextView) :
    def __init__ (self, item) :
        gtk.TextView.__init__(self)
        self._item = item
        self.modify_font(pango.FontDescription("monospace"))
        self.set_border_window_size(gtk.TEXT_WINDOW_LEFT, 20)
        self.connect("expose_event", self._line_numbers_expose)
        self.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse("white"))
        self.set_property("wrap-mode", "word")
    def _get_lines(self, first_y, last_y, buffer_coords, numbers):
        # Get iter at first y
        iter, top = self.get_line_at_y(first_y)
        # For each iter, get its location and add it to the arrays.
        # Stop when we pass last_y
        count = 0
        size = 0
        while not iter.is_end():
            y, height = self.get_line_yrange(iter)
            buffer_coords.append(y)
            line_num = iter.get_line()
            numbers.append(line_num)
            count += 1
            if (y + height) >= last_y:
                break
            iter.forward_line()
        return count
    def _line_numbers_expose(self, widget, event, user_data=None):
        # See if this expose is on the line numbers window
        left_win = widget.get_window(gtk.TEXT_WINDOW_LEFT)
        if event.window == left_win:
            type = gtk.TEXT_WINDOW_LEFT
            target = left_win
        else:
            return False
        first_y = event.area.y
        last_y = first_y + event.area.height
        x, first_y = widget.window_to_buffer_coords(type, 0, first_y)
        x, last_y = widget.window_to_buffer_coords(type, 0, last_y)
        numbers = []
        pixels = []
        count = self._get_lines(first_y, last_y, pixels, numbers)
        # Draw fully internationalized numbers!
        layout = widget.create_pango_layout("")
        for i in range(count):
            x, pos = widget.buffer_to_window_coords(type, 0, pixels[i])
            layout.set_text(_(self._item))
            widget.style.paint_layout(target, widget.state, False,
                                      None, widget, None, 4, pos, layout)
        # don't stop emission, need to draw children
        return False
    def get_text (self) :
        buff = self.get_buffer()
        return buff.get_text(buff.get_start_iter(), buff.get_end_iter())
    def clear (self) :
        buff = self.get_buffer()
        buff.delete(buff.get_start_iter(), buff.get_end_iter())
    def append (self, text) :
        buff = self.get_buffer()
        if self.get_text() != "" :
            end = buff.get_end_iter()
            if end.backward_char() and end.get_char() != "\n" :
                buff.insert(buff.get_end_iter(), "\n")
        buff.insert(buff.get_end_iter(), _(text))
    def __iter__ (self) :
        return ((self._item, line.rstrip())
                for line in self.get_text().split("\n")
                if line.strip() != "")

class BzrLog (pygg.AppWindow) :
    __glade__ = open("bzrlog.glade").read()
    def _quit (self) :
        self._save()
        self.quit()
    def _save (self) :
        self.Todo.save()
        self._save_news()
    def _save_news (self) :
        news = ItemList(self._categories, True)
        news.load("NEWS")
        for item, text in self._get_items() :
            news.add_item(item, text)
        news.save("NEWS")
    def _build (self) :
        pygg.AppWindow._build(self)
        self.Files = FileManager()
        self.ScrolledwindowFiles.add(self.Files)
        for signal in ("move-cursor", "row-collapsed", "row-expanded",
                       "select-cursor-parent", "unselect-all",
                       "button-release-event") :
            self.Files.connect(signal, self.on_Files_selection_changed)
        self.Todo = TodoList()
        self.ScrolledwindowTodo.add(self.Todo)
        for signal in ("move-cursor", "row-collapsed", "row-expanded",
                       "select-cursor-parent", "unselect-all",
                       "button-release-event") :
            self.Todo.connect(signal, self.on_Todo_selection_changed)
    def _load (self) :
        self._wt, _ = bzrlib.workingtree.WorkingTree.open_containing()
        os.chdir(self._wt.abspath(""))
        self._br = self._wt.branch
        self._load_prefs()
        self._refresh_project()
        self.Todo.load(self._categories)
        self._pages = {}
        self._pagenum = {}
        for item, name in self._categories :
            self._pages[name] = CategoryTextView(item)
            self._pagenum[name] = self.notebook.get_n_pages()
            self.notebook.append_page(self._pages[name], gtk.Label(name))
        self.notebook.show_all()
        self._load_entries()
    def _load_prefs (self) :
        if os.path.isfile(".bzrlog") :
            prefs = open(".bzrlog")
        elif os.path.isfile(os.path.expanduser("~/.bzrlog")) :
            prefs = open(os.path.expanduser("~/.bzrlog"))
        else :
            prefs = StringIO.StringIO("cat=+Feature\n"
                                      "cat=!Bug\n"
                                      "cat=*Change\n")
        self._categories = CategorySet()
        for line in (l.strip() for l in prefs if not l.startswith("#")) :
            if line == "" :
                break
            opt, val = line.split("=", 1)
            if opt == "tab" :
                self.notebook.set_current_page(int(val))
            elif opt == "win" :
                self.parse_geometry(val)
            elif opt == "cat" :
                self._categories.add(val[0], val[1:])
        prefs.close()
    def _load_entries (self) :
        try :
            prefs = open(".bzrlog")
        except :
            return
        for line in prefs :
            if line.strip() == "" :
                break
        for line in (l.strip() for l in prefs if l.strip() != "") :
            item, name = self._categories[line[0]]
            self._pages[name].append(line[1:])
        prefs.close()
    def _refresh_project (self) :
        self.Files.load(self._wt, self.ButtonShowFiles.get_active())
        self.EntryName.set_text(str(self._br.nick))
        self.ButtonSaveName.set_sensitive(False)
        revision = self._br.last_revision()
        self.EntryVersion.set_text(self._get_version())
        self.ButtonSaveVersion.set_sensitive(False)
    def _get_version (self) :
        revno = self._br.revno()
        version = "rev. %s" % revno
        _version = None
        for tag in self._br.tags.get_tag_dict() :
            if tag.startswith("version-") :
                try :
                    v = tuple(int(i) for i in tag[8:].split("."))
                except :
                    continue
                if _version < v :
                    version = tag[8:]
                    _ver = v
        return version
    def _get_items (self) :
        for item, name in self._categories :
            tv = self._pages[name]
            for entry in tv :
                yield entry
    def _get_text (self) :
        lst = ItemList(self._categories)
        for item, name in self._get_items() :
            lst.add_item(item, name)
        return lst.format()
    def _clear (self) :
        for item, name in self._categories :
            self._pages[name].clear()
    def run (self) :
        self._load()
        pygg.AppWindow.run(self)
    def on_Files_selection_changed (self, *args) :
        selection = self.Files.get_file()
        if selection is None :
            for button in (self.ButtonAddFile, self.ButtonRemoveFile,
                           self.ButtonDeleteFile) :
                button.set_sensitive(False)
        else :
            path, isdir, status, ident = selection
            hasid = status not in ("?", "removed")
            if str(self._wt.relpath(path)) == "" :
                self.ButtonAddFile.set_sensitive(False)
                self.ButtonRemoveFile.set_sensitive(False)
                self.ButtonDeleteFile.set_sensitive(False)
            else :
                self.ButtonAddFile.set_sensitive(not hasid)
                self.ButtonRemoveFile.set_sensitive(hasid)
                self.ButtonDeleteFile.set_sensitive(True)
    def on_BzrLog_delete_event (self, widget, event) :
        self._quit()
    def on_ButtonQuit_clicked (self, toolbutton) :
        self._quit()
    def on_ButtonCommit_clicked (self, toolbutton) :
        self.Todo.save()
        news = ItemList(self._categories, True)
        news.load("NEWS")
        self._save_news()
        dialog = gtk_commit.CommitDialog(self._wt, self._wt.abspath(""))
        self._tv = None
        def search (widget) :
            if isinstance(widget, gtk.TextView) and widget.is_focus() :
                self._tv = widget
            else :
                try :
                    widget.forall(search)
                except :
                    pass
        dialog.forall(search)
        self._tv.get_buffer().set_text(self._get_text())
        dialog.set_title("Commit...")
        response = dialog.run()
        if response == gtk.RESPONSE_OK :
            self._clear()
        else :
            news.save("NEWS")
        dialog.destroy()
        self._refresh_project()
    def on_ButtonExport_clicked (self, toolbutton) :
        print "on_ButtonExport_clicked"
    def on_ButtonConfig_clicked (self, toolbutton) :
        print "on_ButtonConfig_clicked"
    def on_ButtonAbout_clicked (self, toolbutton) :
        print "on_ButtonAbout_clicked"
    def on_ButtonAddFile_clicked (self, toolbutton) :
        try :
            path = self.Files.get_file()[0]
        except :
            return
        self._wt.smart_add([self._wt.relpath(path)])
        self._refresh_project()
    def on_ButtonRemoveFile_clicked (self, toolbutton) :
        try :
            path = self.Files.get_file()[0]
        except :
            return
        self._wt.remove(self._wt.relpath(path))
        self._refresh_project()
    def on_ButtonDeleteFile_clicked (self, toolbutton) :
        try :
            path, _, status, _ = self.Files.get_file()
        except :
            return
        if status not in ("?", "removed") :
            self._wt.remove(self._wt.relpath(path))
        try :
            if os.path.isdir(path) :
                shutil.rmtree(path)
            else :
                os.unlink(path)
        except :
            pass
        self._refresh_project()
    def on_ButtonRefreshFiles_clicked (self, toolbutton) :
        self._refresh_project()
    def on_Todo_selection_changed (self, *args) :
        sensitive = self.Todo.get_item() is not None
        for button in (self.ButtonDeleteTodo, self.ButtonDoneTodo) :
            button.set_sensitive(sensitive)
    def on_ButtonNewTodo_clicked (self, toolbutton) :
        self.Todo.add_item(self._categories.default(), "new task...")
    def on_ButtonDeleteTodo_clicked (self, toolbutton) :
        self.Todo.del_item()
    def on_ButtonDoneTodo_clicked (self, toolbutton) :
        item, text = self.Todo.get_item()
        _, name = self._categories[item]
        self._pages[name].append(text)
        self.notebook.set_current_page(self._pagenum[name])
        self.Todo.del_item()
    def on_EntryName_changed (self, editable) :
        if self.EntryName.get_text().strip() != "" :
            self.ButtonSaveName.set_sensitive(True)
    def on_EntryVersion_changed (self, editable) :
        if self.EntryVersion.get_text().strip() != "" :
            self.ButtonSaveVersion.set_sensitive(True)
    def on_ButtonSaveName_clicked (self, button) :
        name = self.EntryName.get_text().strip()
        if name != "" :
            self._br.nick = name
        self.ButtonSaveName.set_sensitive(False)
        self._refresh_project()
    def on_ButtonSaveVersion_clicked (self, button) :
        new = self.EntryVersion.get_text().strip()
        try :
            version = tuple(int(i) for i in new.split("."))
        except :
            return
        old = self._get_version()
        for tag in self._br.tags.get_tag_dict() :
            if tag.startswith("version-") :
                self._br.tags.delete_tag(tag)
        self._br.tags.set_tag("version-" + new, self._br.last_revision())
        try :
            outfile = open("VERSION", "w")
            print >>outfile, new
            outfile.close()
        except :
            pass
        self.ButtonSaveVersion.set_sensitive(False)
        self.Todo.save()
        news = ItemList(self._categories, True)
        news.load("NEWS")
        news.add_text("version %s" % old)
        news.add_text("")
        news.save("NEWS")
        self._wt.commit(message="version %s" % new,
                        specific_files=["VERSION", "TODO", "NEWS"])
        self._refresh_project()

if __name__ == "__main__" :
    BzrLog().run()
